HDF5_EXTENSION = "hdf5"
HDF_SEPARATOR = "/"
MODEL_SUBFOLDER = "MODEL"
CLASS_WEIGHT_FILE = "class_weights.json"
TRAIN_DS_NAME = "train"
VALIDATION_DS_NAME = "validation"
TEST_DS_NAME = "test"

# Used to determine whether a dataset can safely fit into
# The available memory
MEM_SAFETY_FACTOR = 1.05

# this constant is used everywhere to check whether
# we should automagically apply that thing
AUTO = "AUTO"


FEATURE_INDEX = "feature"
LABEL_INDEX = "label"
